import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import networkx as nx
from simsnn.core.detectors import Raster, Multimeter


class Simulator:
    """Simulator

    Parameters
    ----------
    network : Network
        Network to simulate
    detectors : List
        List of detectors
    """

    def __init__(self, network, seed=None):
        self.network = network
        self.multimeter = Multimeter()
        self.raster = Raster()
        if seed != None:
            self.network.update_rng(np.random.RandomState(seed))

    def run(self, steps, plotting=False, options=None):
        """Run the simulator

        Parameters
        ----------
        steps : int
            Number of steps to simulate
        """
        options = {} if options is None else options
        self.raster.initialize(steps)
        self.multimeter.initialize(steps)

        for _ in range(steps):
            self.network.step()
            self.raster.step()
            self.multimeter.step()

        if plotting:
            self.print_detectors(steps, options)

    def to_inet_string(self):
        inet_str = ""
        inet_str += self.raster.to_inet_string() + "\n\n"
        inet_str += self.multimeter.to_inet_string() + "\n\n"
        return inet_str

    def print_detectors(self, steps, options):
        rasterdata = self.raster.get_measurements()
        print("Rasterdata:")
        print(rasterdata.T)
        multimeterdata = self.multimeter.get_measurements()
        print("\nMultimeterdata:")
        print(multimeterdata.T)

        graph = options.get("graph", True)
        ntd = len(rasterdata.T)
        nvd = len(multimeterdata.T)

        ntd_pos = 1 if graph else 0
        nvd_offset = ntd_pos
        nvd_offset += 1 if ntd else 0

        fig, _ = plt.subplots(
            constrained_layout=True, nrows=nvd_offset + nvd, figsize=(14, 8)
        )

        options = {
            "with_labels": True,
            "node_color": "white",
            "edgecolors": "blue",
            "ax": fig.axes[0],
            "node_size": 1100,
            "pos": nx.circular_layout(self.network.graph),
        }
        if graph:
            nx.draw_networkx(self.network.graph, **options)

        if ntd:
            fig.axes[ntd_pos].matshow(rasterdata.T, cmap="gray", aspect="auto")
            fig.axes[ntd_pos].set_xticks(np.arange(-0.5, steps, 1), minor=True)
            fig.axes[ntd_pos].grid(
                which="minor", color="gray", linestyle="-", linewidth=2
            )
            fig.axes[ntd_pos].set_yticks(np.arange(-0.5, ntd, 1), minor=True)
            fig.axes[ntd_pos].xaxis.set_major_locator(ticker.MultipleLocator(1))
            fig.axes[ntd_pos].set_yticks(np.arange(0, ntd, 1))
            fig.axes[ntd_pos].set_yticklabels([t.ID for t in self.raster.targets])

        for i in range(nvd):
            fig.axes[i + nvd_offset].plot(multimeterdata[:, i])
            fig.axes[i + nvd_offset].set_ylim(top=(max(multimeterdata.T[i]) + 0.5))
            fig.axes[i + nvd_offset].set_ylabel(self.multimeter.targets[i].ID)
            fig.axes[i + nvd_offset].grid(which="major")
            fig.axes[i + nvd_offset].xaxis.set_major_locator(ticker.MultipleLocator(1))

        plt.show()



import numpy as np


class Raster:
    def __init__(self, targets=None, ID=None, increment_count=True):
        self.targets = targets if targets is not None else []
        self.ID = ID

    def initialize(self, steps):
        self.spikes = np.zeros((steps, len(self.targets)), dtype=bool)
        self.index = 0

    def step(self):
        self.spikes[self.index, :] = [target.out > 0 for target in self.targets]
        self.index += 1

    def get_measurements(self):
        return self.spikes

    def get_labels(self):
        return [t.ID for t in self.targets]

    def addTarget(self, target):
        if isinstance(target, list):
            self.targets.extend(target)
        else:
            self.targets.append(target)

    def addTargets(self, target):
        self.targets.extend(target)

    def to_inet_string(self):
        inet_string = (
            self.__class__.__name__ + "_" + str(self.ID) + " = "
            "simulator.create" + self.__class__.__name__ + "()"
        )

        for t in self.targets:
            inet_string += (
                self.__class__.__name__
                + "_"
                + str(self.ID)
                + ".addTarget("
                + t.__class__.__name__
                + "_"
                + str(t.ID)
                + ")\n"
            )

        return inet_string


class Multimeter:
    def __init__(self, targets=None, ID=None, increment_count=True):
        self.targets = targets if targets is not None else []
        self.ID = ID

    def initialize(self, steps):
        self.V = np.zeros((steps, len(self.targets)))
        self.index = 0

    def step(self):
        self.V[self.index, :] = [target.V for target in self.targets]
        self.index += 1

    def get_measurements(self):
        return self.V

    def get_labels(self):
        return [t.ID for t in self.targets]

    def addTarget(self, target):
        if isinstance(target, list):
            self.targets.extend(target)
        else:
            self.targets.append(target)

    def addTargets(self, target):
        self.targets.extend(target)

    def to_inet_string(self):
        inet_string = (
            self.__class__.__name__ + "_" + str(self.ID) + " = "
            "simulator.create" + self.__class__.__name__ + "()\n"
        )

        for t in self.targets:
            inet_string += (
                self.__class__.__name__
                + "_"
                + str(self.ID)
                + ".addTarget("
                + t.__class__.__name__
                + "_"
                + str(t.ID)
                + ")\n"
            )

        return inet_string



import argparse
from simsnn.examples import circuit


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d", default=5, metavar="N", type=int, help="The amount of steps to simulate"
    )
    parser.add_argument(
        "-f",
        default="circuit",
        metavar="F",
        type=str,
        help="The example network to run",
    )
    parser.add_argument(
        "-g",
        "--graph",
        dest="graph",
        action="store_false",
        help="Suppress the plotting of a graph",
    )
    parser.set_defaults(graph=True)

    args = parser.parse_args()

    if args.f == "circuit":
        circuit.run(duration=args.d, options={"graph": args.graph})


if __name__ == "__main__":
    main()



import networkx as nx

from simsnn.core.nodes import LIF, InputTrain, RandomSpiker
from simsnn.core.connections import Synapse


class Network:
    """Network containing a list of nodes and synapses

    Parameters
    ----------
    nodes : list
        List of nodes in the network
    synapses : list
        List of synapses connecting nodes
    """

    def __init__(self, nodes=None, synapses=None):
        self.nodes = nodes if nodes is not None else []
        self.synapses = synapses if synapses is not None else []
        self.graph = nx.DiGraph()

    def createLIF(
        self,
        m=1.0,
        V_init=0,
        V_reset=0,
        V_min=0,
        thr=1,
        amplitude=1,
        I_e=0,
        noise=0,
        rng=None,
        ID=None,
        increment_count=True,
    ):
        self.graph.add_node(ID)
        node = LIF(
            m,
            V_init,
            V_reset,
            V_min,
            thr,
            amplitude,
            I_e,
            noise,
            rng,
            ID,
            increment_count,
        )
        self.nodes.append(node)
        return node

    def createInputTrain(self, train, loop, ID=None, increment_count=True):
        self.graph.add_node(ID)
        node = InputTrain(train, loop, ID, increment_count)
        self.nodes.append(node)
        return node

    def createRandomSpiker(
        self, p, amplitude=1, rng=None, ID=None, increment_count=True
    ):
        node = RandomSpiker(p, amplitude, rng, ID, increment_count)
        self.nodes.append(node)
        return node

    def createSynapse(self, pre, post, w=1.0, d=1, ID=None, increment_count=True):
        self.graph.add_edge(pre.ID, post.ID)
        synapse = Synapse(pre, post, w, d, ID, increment_count)
        self.synapses.append(synapse)
        return synapse

    def step(self):
        for node in self.nodes:  # update all nodes
            node.step()
        for synapse in self.synapses:  # update all synapses
            synapse.step()

    def update_rng(self, rng):
        for node in self.nodes:
            node.update_rng(rng)

    def to_inet_string(self):
        inet_str = ""

        for n in self.nodes:
            inet_str += n.to_inet_string() + "\n"

        inet_str += "\n"

        for s in self.synapses:
            inet_str += s.to_inet_string() + "\n"

        return inet_str



from simsnn.core.networks import Network
from simsnn.core.simulators import Simulator


def run(duration=10, options=None):
    options = {} if options is None else options

    # Create the network and the simulator object
    net = Network()
    sim = Simulator(net)

    # Create a programmed neuron, that spikes on times 1 and 3,
    # does not repeat it's programming and has the ID "pn".
    programmed_neuron = net.createInputTrain(train=[0, 1, 0, 1], loop=False, ID="pn")

    # Create a LIF neuron, with a membrane voltage threshold of 1,
    # a post spike reset value of 0 and no voltage decay (m=1).
    lif_neuron = net.createLIF(ID="ln", thr=1, V_reset=0, m=1)

    # Create a Synapse, between the programmed neuron and the LIF neuron,
    # with a voltage weight of 1 and a delay of 1.
    net.createSynapse(pre=programmed_neuron, post=lif_neuron, ID="pn-ln", w=1, d=1)

    # Add all neurons to the raster
    sim.raster.addTarget([programmed_neuron, lif_neuron])
    # Add all neurons to the multimeter
    sim.multimeter.addTarget([programmed_neuron, lif_neuron])

    # Run the simulation for 10 rounds, enable the plotting of the raster,
    # the multimeter and the network structure.
    sim.run(duration, plotting=True, options=options)



import numpy as np

"""
neurons
"""


class AbstractNeuron:
    """Abstract class for a neuron

    Attributes
    ----------
    I : float (Default: 0)
        Current membrane voltage of the neuron
    out : float (Default: 0)
        Current neuron output
    """

    def __init__(self, amplitude=1):
        self.amplitude = amplitude
        self.I = 0
        self.out = 0
        self.V = 0

    def update_rng(self, rng):
        if hasattr(self, "rng"):
            self.rng = rng


class LIF(AbstractNeuron):
    """Leaky integrate-and-fire based on the Sandia model

    Parameters
    ----------
    m : float
        Leakage constant
    V_init : float
        Initial membrane voltage (Default: 0)
    V_reset : float
        Reset voltage when the neuron has spiked (Default: 0)
    thr : float
        Spiking threshold (Default: 1)
    amplitude : float (Default: 1)
        Amplitude of the output when the neuron spikes
    I_e : float (Default: 0)
        Constant input current
    noise : float (Default: 0)
        Standard deviation of the normal distribution that is sampled from
        to add noise to the membrane voltage at each step
    rng : np.random.RandomState
        Random generator for the noise
    """

    count = 0

    def __init__(
        self,
        m=1.0,
        V_init=0,
        V_reset=0,
        V_min=0,
        thr=1,
        amplitude=1,
        I_e=0,
        noise=0,
        rng=None,
        ID=None,
        increment_count=True,
    ):
        AbstractNeuron.__init__(self, amplitude)
        self.m = m
        self.V = V_init
        self.V_reset = V_reset
        self.V_min = V_min
        self.thr = thr
        self.I_e = I_e
        self.I = I_e
        self.rng = rng if rng is not None else np.random.RandomState()
        self.noise = noise

        if ID is None:
            self.ID = LIF.count + 1
        else:
            self.ID = ID
        if increment_count:
            LIF.count += 1

    def step(self):
        self.V = self.V * self.m + self.I  # update V
        if self.noise > 0:
            self.V += self.rng.normal(scale=self.noise)  # add noise
        self.V = max(self.V_min, self.V)
        self.I = self.I_e  # reset I with I_e
        if self.V >= self.thr:  # check for spike
            self.V = self.V_reset
            self.out = self.amplitude
        else:
            self.out = 0

    def to_inet_string(self):
        return (
            f"{self.__class__.__name__}_{str(self.ID)}"
            f" = network.create {self.__class__.__name__}"
            f"({str(self.m)}, {str(self.V)}, {str(self.V_reset)}, "
            f"{str(self.V_min)}, {str(self.thr)}, {str(self.amplitude)}, "
            f"{str(self.I_e)}, {str(self.noise)}"
        )

    def to_string_wo_id(self):
        return (
            "classname=" + self.__class__.__name__ + "_" + " = "
            "network.create"
            + self.__class__.__name__
            + "("
            + "m="
            + str(self.m)
            + ", "
            + "V="
            + str(self.V)
            + ", "
            + "V_reset="
            + str(self.V_reset)
            + ", "
            + "V_min="
            + str(self.V_min)
            + ", "
            + "thr="
            + str(self.thr)
            + ", "
            + "amplitude="
            + str(self.amplitude)
            + ", "
            + "I_e="
            + str(self.I_e)
            + ", "
            + "noise="
            + str(self.noise)
            + ")"
        )

    def print_neuron_status(self):
        print(f"ID={self.ID}")
        print(f"amplitude={self.amplitude}")
        print(f"I={self.I}")
        print(f"out={self.out}")
        print(f"V={self.V}")
        print(f"m = {self.m}")
        print(f"V_reset = {self.V_reset}")
        print(f"V_min = {self.V_min}")
        print(f"thr = {self.thr}")
        print(f"I_e = {self.I_e}")
        print(f"rng = {self.rng }")
        print(f"noise ={self. noise}\n")


"""
Generators
"""


class InputTrain(AbstractNeuron):
    """Generator that outputs a given train of events

    Parameters
    ----------
    train : array_like
        Output train
    loop : boolean
        If true, the train will be looped. Otherwise, the output is 0 at the end of the train.
    """

    count = 0

    def __init__(self, train, loop, ID=None, increment_count=True):
        AbstractNeuron.__init__(self, 1)
        self.train = train
        self.loop = loop
        self.size = len(train)
        self.index = 0

        if ID is None:
            self.ID = InputTrain.count + 1
        else:
            self.ID = ID
        if increment_count:
            InputTrain.count += 1

    def step(self):
        if self.index >= self.size:
            if self.loop:
                self.index = 0
                self.V = self.train[self.index]
            else:
                self.V = 0

        else:
            self.V = self.train[self.index]

        self.out = self.V
        self.index += 1

    def to_inet_string(self):
        return (
            self.__class__.__name__ + "_" + str(self.ID) + " = "
            "network.create"
            + self.__class__.__name__
            + "("
            + str(self.train)
            + ", "
            + str(self.loop)
            + ")"
        )


'''class PoissonGenerator(AbstractNeuron):
    """Generator that fires with Poisson statistics, i.e. exponentially
    distributed interspike intervals.

    Parameters
    ----------
    I_e : float (Default: 0)
        Constant input current
    amplitude : float (Default: 1)
        Amplitude of the output
    rng : np.random.RandomState
        Random generator
    """

    def __init__(self, I_e=0, amplitude=1, rng=None):
        AbstractNeuron.__init__(self, amplitude)
        self.I_e = I_e
        self.rng = rng if rng != None else np.random.RandomState()

    def step(self):
        self.V = 0
        if self.I > 0:
            while (-np.log(1-self.rng.rand()) / self.I) < 1:
                self.V += 1

        self.out = self.V * self.amplitude
        self.I = self.I_e
'''


class RandomSpiker(AbstractNeuron):
    """Generator that fires with a given probability at each time step

    Parameters
    ----------
    I_e : float (Default: 0)
        Constant input current
    amplitude : float (Default: 1)
        Amplitude of the output
    rng : np.random.RandomState
        Random generator
    """

    count = 0

    def __init__(self, p, amplitude=1, rng=None, ID=None, increment_count=True):
        AbstractNeuron.__init__(self, amplitude)
        self.p = p
        self.rng = rng if rng is not None else np.random.default_rng()

        if ID is None:
            self.ID = RandomSpiker.count + 1
        else:
            self.ID = ID
        if increment_count:
            RandomSpiker.count += 1

    def step(self):
        self.V = 0
        if self.rng.random() < self.p:
            self.V = self.amplitude

        self.out = self.V

    def to_inet_string(self):
        return (
            self.__class__.__name__ + "_" + str(self.ID) + " = "
            "network.create"
            + self.__class__.__name__
            + "("
            + str(self.p)
            + ", "
            + str(self.amplitude)
            + ")"
        )



import numpy as np


class Synapse:
    """Connection between two neurons

    Parameters
    ----------
    pre : Neuron
        Presynaptic neuron
    post : Neuron
        Postsynaptic neuron
    w : float
        Connection weight
    d : int
        Synaptic delay (number of timesteps)
    """

    count = 0

    def __init__(self, pre, post, w=1, d=1, ID=None, increment_count=True):
        if d < 1:
            raise ValueError("Synaptic delay must be at least 1")
        self.pre = pre
        self.post = post
        self.w = w
        self.out_pre = np.zeros(
            (d)
        )  # store output of the presynaptic neuron during d timesteps
        self.index = 0

        if ID is None:
            self.ID = Synapse.count + 1
        else:
            self.ID = ID
        if increment_count:
            Synapse.count += 1

    def step(self):

        self.out_pre[self.index] = self.pre.out  # store current output of pre
        self.index = (self.index + 1) % len(self.out_pre)
        self.post.I += self.w * self.out_pre[self.index]  # add w*pre_{t-d} to post

    def to_inet_string(self):
        return (
            "network.create"
            + self.__class__.__name__
            + "("
            + self.pre.__class__.__name__
            + "_"
            + str(self.pre.ID)
            + ", "
            + self.post.__class__.__name__
            + "_"
            + str(self.pre.ID)
            + ", "
            + str(self.w)
            + ", "
            + str(self.out_pre.shape[0])
            + ")"
        )
